@@ -2,7 +2,7 @@
  * DMI Decode
  *
  *   (C) 2000-2002 Alan Cox <alan@redhat.com>
- *   (C) 2002-2005 Jean Delvare <khali@linux-fr.org>
+ *   (C) 2002-2007 Jean Delvare <khali@linux-fr.org>
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  *   are deemed to be part of the source code.
  *
  * Unless specified otherwise, all references are aimed at the "System
- * Management BIOS Reference Specification, Version 2.4" document,
+ * Management BIOS Reference Specification, Version 2.5" document,
  * available from http://www.dmtf.org/standards/smbios/.
  *
  * Note to contributors:
@@ -33,7 +33,7 @@
  * information does not come from the above mentioned specification.
  *
  * Additional references:
- *  - Intel AP-485 revision 28
+ *  - Intel AP-485 revision 31
  *    "Intel Processor Identification and the CPUID Instruction"
  *    http://developer.intel.com/design/xeon/applnots/241618.htm
  *  - DMTF Master MIF version 040707
@@ -42,9 +42,9 @@
  *  - IPMI 2.0 revision 1.0
  *    "Intelligent Platform Management Interface Specification"
  *    http://developer.intel.com/design/servers/ipmi/spec.htm
- *  - AMD publication #20734 revision 3.04
- *    "AMD Processor Recognition Application Note"
- *    http://www.amd.com/us-en/assets/content_type/white_papers_and_tech_docs/20734.pdf
+ *  - AMD publication #25481 revision 2.18
+ *    "CPUID Specification"
+ *    http://www.amd.com/us-en/assets/content_type/white_papers_and_tech_docs/25481.pdf
  */
 
 #include <stdio.h>
@@ -52,41 +52,24 @@
 #include <stdlib.h>
 #include <unistd.h>
 
-#ifdef __ia64__
-#define USE_EFI
-#endif /* __ia64__ */
-
 #include "version.h"
 #include "config.h"
 #include "types.h"
 #include "util.h"
 #include "dmidecode.h"
 #include "dmiopt.h"
+#include "dmioem.h"
 
 static const char *out_of_spec = "<OUT OF SPEC>";
 static const char *bad_index = "<BAD INDEX>";
 
-struct dmi_header
-{
-	u8 type;
-	u8 length;
-	u16 handle;
-};
-
-#ifdef ALIGNMENT_WORKAROUND
-#define HANDLE(x) WORD((u8 *)&(x->handle))
-#else
-#define HANDLE(x) x->handle
-#endif
-
-
 /*
  * Type-independant Stuff
  */
 
-static const char *dmi_string(struct dmi_header *dm, u8 s)
+const char *dmi_string(struct dmi_header *dm, u8 s)
 {
-	char *bp=(char *)dm;
+	char *bp=(char *)dm->data;
 	size_t i, len;
 
 	if(s==0)
@@ -181,11 +164,11 @@ static void dmi_dump(struct dmi_header *h, const char *prefix)
 	{
 		printf("%s\t", prefix);
 		for(i=0; i<16 && i<h->length-(row<<4); i++)
-			printf("%s%02X", i?" ":"", ((u8 *)h)[(row<<4)+i]);
+			printf("%s%02X", i?" ":"", (h->data)[(row<<4)+i]);
 		printf("\n");
 	}
 
-	if(((u8 *)h)[h->length] || ((u8 *)h)[h->length+1])
+	if((h->data)[h->length] || (h->data)[h->length+1])
 	{
 		printf("%sStrings:\n", prefix);
 		i=1;
@@ -458,10 +441,12 @@ const char *dmi_chassis_type(u8 code)
 		"RAID Chassis",
 		"Rack Mount Chassis",
 		"Sealed-case PC",
-		"Multi-system" /* 0x19 */
+		"Multi-system",
+	   	"CompactPCI",
+		"AdvancedTCA" /* 0x1B */
 	};
 	
-	if(code>=0x01 && code<=0x19)
+	if(code>=0x01 && code<=0x1B)
 		return type[code-0x01];
 	return out_of_spec;
 }
@@ -707,10 +692,10 @@ const char *dmi_processor_family(u8 code)
 		"Athlon 64",
 		"Opteron",
 		"Sempron",
-		NULL, /* 0x86 */
-		NULL,
-		NULL,
-		NULL,
+		"Turion 64", 
+		"Dual-Core Opteron",
+		"Athlon 64 X2",
+		NULL, /* 0x89 */
 		NULL,
 		NULL,
 		NULL,
@@ -759,10 +744,10 @@ const char *dmi_processor_family(u8 code)
 		"Athlon MP",
 		"Itanium 2",
 		"Pentium M",
-		NULL, /* 0xBA */
-		NULL,
-		NULL,
-		NULL,
+		"Celeron D",
+		"Pentium D",
+		"Pentium EE",
+		NULL, /* 0xBD */
 		NULL,
 		NULL,
 		NULL,
@@ -839,7 +824,7 @@ const char *dmi_processor_family(u8 code)
 
 static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *prefix)
 {
-	/* Intel AP-485 revision 28, table 5 */
+	/* Intel AP-485 revision 31, table 3-4 */
 	static const char *flags[32]={
 		"FPU (Floating-point unit on-chip)", /* 0 */
 		"VME (Virtual mode extension)",
@@ -871,12 +856,12 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 		"SS (Self-snoop)",
 		"HTT (Hyper-threading technology)",
 		"TM (Thermal monitor supported)",
-		NULL, /* 30 */
+		"IA64 (IA64 capabilities)",
 		"PBE (Pending break enabled)" /* 31 */
 	};
 	/*
 	 * Extra flags are now returned in the ECX register when one calls
-	 * the CPUID instruction. Their meaning is explained in table 6, but
+	 * the CPUID instruction. Their meaning is explained in table 3-5, but
 	 * DMI doesn't support this yet.
 	 */
 	u32 eax, edx;
@@ -921,12 +906,12 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 	else if((type>=0x0B && type<=0x13) /* Intel, Cyrix */
 	|| (type>=0xB0 && type<=0xB3) /* Intel */
 	|| type==0xB5 /* Intel */
-	|| type==0xB9) /* Intel */
+	|| (type>=0xB9 && type<=0xBC)) /* Intel */
 		sig=1;
 	else if((type>=0x18 && type<=0x1D) /* AMD */
 	|| type==0x1F /* AMD */
 	|| (type>=0xB6 && type<=0xB7) /* AMD */
-	|| (type>=0x83 && type<=0x85)) /* AMD */
+	|| (type>=0x83 && type<=0x88)) /* AMD */
 		sig=2;
 	else if(type==0x01 || type==0x02)
 	{
@@ -952,23 +937,21 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 	{
 		case 1: /* Intel */
 			printf("%sSignature: Type %u, Family %u, Model %u, Stepping %u\n",
-				prefix, (eax>>12)&0x3, ((eax>>16)&0xFF0)+((eax>>8)&0x00F),
+				prefix, (eax>>12)&0x3, ((eax>>20)&0xFF)+((eax>>8)&0x0F),
 				((eax>>12)&0xF0)+((eax>>4)&0x0F), eax&0xF);
 			break;
 		case 2: /* AMD */
-			printf("%sSignature: %s %X, %s %X, Stepping %X\n",
+			printf("%sSignature: Family %u, Model %u, Stepping %u\n",
 				prefix,
-				((eax>>8)&0xF)==0xF?"Extended Family":"Family",
-				((eax>>8)&0xF)==0xF?(eax>>20)&0xFF:(eax>>8)&0xF,
-				((eax>>4)&0xF)==0xF?"Extended Model":"Model",
-				((eax>>4)&0xF)==0xF?(eax>>16)&0xF:(eax>>4)&0xF,
+				((eax>>8)&0xF)+(((eax>>8)&0xF)==0xF?(eax>>20)&0xFF:0),
+				((eax>>4)&0xF)|(((eax>>8)&0xF)==0xF?(eax>>12)&0xF0:0),
 				eax&0xF);
 			break;
 	}
 
 	edx=DWORD(p+4);
 	printf("%sFlags:", prefix);
-	if((edx&0x3FF7FDFF)==0)
+	if((edx&0xFFEFFBFF)==0)
 		printf(" None\n");
 	else
 	{
@@ -1052,10 +1035,13 @@ static const char *dmi_processor_upgrade(u8 code)
 		"Socket 478",
 		"Socket 754",
 		"Socket 940",
-		"Socket 939" /* 0x12 */
+		"Socket 939",
+	   	"Socket mPGA604",
+		"Socket LGA771",
+		"Socket LGA775" /* 0x15 */
 	};
 	
-	if(code>=0x01 && code<=0x11)
+	if(code>=0x01 && code<=0x15)
 		return upgrade[code-0x01];
 	return out_of_spec;
 }
@@ -1073,6 +1059,26 @@ static void dmi_processor_cache(u16 code, const char *level, u16 ver)
 		printf(" 0x%04X", code);
 }
 
+static void dmi_processor_characteristics(u16 code, const char *prefix)
+{
+	/* 3.3.5.9 */
+	static const char *characteristics[]={
+		"64-bit capable" /* 2 */
+	};
+
+	if((code&0x0004)==0)
+		printf(" None\n");
+	else
+	{
+		int i;
+
+		printf("\n");
+		for(i=2; i<=2; i++)
+			if(code&(1<<i))
+				printf("%s%s\n", prefix, characteristics[i-2]);
+	}
+}
+
 /*
  * 3.3.6 Memory Controller Information (Type 5)
  */
@@ -1195,7 +1201,7 @@ static void dmi_memory_module_types(u16 code, const char *sep)
 		"SDRAM" /* 10 */
 	};
 	
-	if((code&0x03FF)==0)
+	if((code&0x07FF)==0)
 		printf(" None");
 	else
 	{
@@ -1306,7 +1312,7 @@ static void dmi_cache_size(u16 code)
 
 static void dmi_cache_types(u16 code, const char *sep)
 {
-	/* 3.3.8.1 */
+	/* 3.3.8.2 */
 	static const char *types[]={
 		"Other", /* 0 */
 		"Unknown",
@@ -1331,7 +1337,7 @@ static void dmi_cache_types(u16 code, const char *sep)
 
 static const char *dmi_cache_ec_type(u8 code)
 {
-	/* 3.3.8.2 */
+	/* 3.3.8.3 */
 	static const char *type[]={
 		"Other", /* 0x01 */
 		"Unknown",
@@ -1348,7 +1354,7 @@ static const char *dmi_cache_ec_type(u8 code)
 
 static const char *dmi_cache_type(u8 code)
 {
-	/* 3.3.8.3 */
+	/* 3.3.8.4 */
 	static const char *type[]={
 		"Other", /* 0x01 */
 		"Unknown",
@@ -1364,7 +1370,7 @@ static const char *dmi_cache_type(u8 code)
 
 static const char *dmi_cache_associativity(u8 code)
 {
-	/* 3.3.8.4 */
+	/* 3.3.8.5 */
 	static const char *type[]={
 		"Other", /* 0x01 */
 		"Unknown",
@@ -1422,7 +1428,8 @@ static const char *dmi_port_connector_type(u8 code)
 		"Mini Centronics Type-26",
 		"Mini Jack (headphones)",
 		"BNC",
-		"IEEE 1394" /* 0x21 */
+		"IEEE 1394",
+	   	"SAS/SATA Plug Receptacle" /* 0x22 */
 	};
 	static const char *type_0xA0[]={
 		"PC-98", /* 0xA0 */
@@ -1432,7 +1439,7 @@ static const char *dmi_port_connector_type(u8 code)
 		"PC-98 Full" /* 0xA4 */
 	};
 	
-	if(code<=0x21)
+	if(code<=0x22)
 		return type[code];
 	if(code>=0xA0 && code<=0xA4)
 		return type_0xA0[code-0xA0];
@@ -1476,14 +1483,16 @@ static const char *dmi_port_type(u8 code)
 		"Video Port",
 		"Audio Port",
 		"Modem Port",
-		"Network Port" /* 0x1F */
+		"Network Port",
+	   	"SATA",
+		"SAS" /* 0x21 */
 	};
 	static const char *type_0xA0[]={
 		"8251 Compatible", /* 0xA0 */
 		"8251 FIFO Compatible" /* 0xA1 */
 	};
 	
-	if(code<=0x1F)
+	if(code<=0x21)
 		return type[code];
 	if(code>=0xA0 && code<=0xA1)
 		return type_0xA0[code-0xA0];
@@ -1607,7 +1616,9 @@ static void dmi_slot_id(u8 code1, u8 code2, u8 type, const char *prefix)
 		case 0x0F: /* AGP */
 		case 0x10: /* AGP */
 		case 0x11: /* AGP */
-		case 0x12: /* PCI */
+		case 0x12: /* PCI-X */
+		case 0x13: /* AGP */
+		case 0xA5: /* PCI Express */
 			printf("%sID: %u\n", prefix, code1);
 			break;
 		case 0x07: /* PCMCIA */
@@ -1637,7 +1648,7 @@ static void dmi_slot_characteristics(u8 code1, u8 code2, const char *prefix)
 	
 	if(code1&(1<<0))
 		printf(" Unknown\n");
-	else if((code1&0x7F)==0 && (code2&0x07)==0)
+	else if((code1&0xFE)==0 && (code2&0x07)==0)
 		printf(" None\n");
 	else
 	{
@@ -1667,17 +1678,20 @@ static const char *dmi_on_board_devices_type(u8 code)
 		"SCSI Controller",
 		"Ethernet",
 		"Token Ring",
-		"Sound" /* 0x07 */
+		"Sound",
+		"PATA Controller",
+		"SATA Controller",
+		"SAS Controller" /* 0x0A */
 	};
 	
-	if(code>=0x01 && code <=0x07)
+	if(code>=0x01 && code<=0x0A)
 		return type[code-0x01];
 	return out_of_spec;
 }
 
 static void dmi_on_board_devices(struct dmi_header *h, const char *prefix)
 {
-	u8 *p=(u8 *)h+4;
+	u8 *p=h->data+4;
 	u8 count=(h->length-0x04)/2;
 	int i;
 
@@ -1704,7 +1718,7 @@ static void dmi_on_board_devices(struct dmi_header *h, const char *prefix)
 
 static void dmi_oem_strings(struct dmi_header *h, const char *prefix)
 {
-	u8 *p=(u8 *)h+4;
+	u8 *p=h->data+4;
 	u8 count=p[0x00];
 	int i;
 	
@@ -1719,7 +1733,7 @@ static void dmi_oem_strings(struct dmi_header *h, const char *prefix)
 
 static void dmi_system_configuration_options(struct dmi_header *h, const char *prefix)
 {
-	u8 *p=(u8 *)h+4;
+	u8 *p=h->data+4;
 	u8 count=p[0x00];
 	int i;
 	
@@ -1734,7 +1748,7 @@ static void dmi_system_configuration_options(struct dmi_header *h, const char *p
 
 static void dmi_bios_languages(struct dmi_header *h, const char *prefix)
 {
-	u8 *p=(u8 *)h+4;
+	u8 *p=h->data+4;
 	u8 count=p[0x00];
 	int i;
 	
@@ -1792,7 +1806,7 @@ static void dmi_event_log_status(u8 code)
 	};
 	
 	printf(" %s, %s",
-		valid[code&(1<<0)], full[code&(1<<1)]);
+		valid[(code>>0)&1], full[(code>>1)&1]);
 }
 
 static void dmi_event_log_address(u8 method, u8 *p)
@@ -2048,10 +2062,11 @@ static const char *dmi_memory_device_form_factor(u8 code)
 		"Row Of Chips",
 		"RIMM",
 		"SODIMM",
-		"SRIMM" /* 0x0E */
+		"SRIMM",
+	   	"FB-DIMM" /* 0x0F */
 	};
 	
-	if(code>=0x01 && code<=0x0E)
+	if(code>=0x01 && code<=0x0F)
 		return form_factor[code-0x01];
 	return out_of_spec;
 }
@@ -2088,10 +2103,11 @@ static const char *dmi_memory_device_type(u8 code)
 		"SGRAM",
 		"RDRAM",
 		"DDR",
-		"DDR2" /* 0x13 */
+		"DDR2",
+	   	"DDR2 FB-DIMM" /* 0x14 */
 	};
 	
-	if(code>=0x01 && code<=0x13)
+	if(code>=0x01 && code<=0x14)
 		return type[code-0x01];
 	return out_of_spec;
 }
@@ -2705,8 +2721,8 @@ static const char *dmi_memory_channel_type(u8 code)
 	static const char *type[]={
 		"Other", /* 0x01 */
 		"Unknown",
-		"RAMBus",
-		"Synclink" /* 0x04 */
+		"RamBus",
+		"SyncLink" /* 0x04 */
 	};
 	
 	if(code>=0x01 && code<=0x04)
@@ -2844,12 +2860,12 @@ static const char *dmi_power_supply_range_switching(u8 code)
  * Main
  */
 
-static void dmi_decode(u8 *data, u16 ver)
+static void dmi_decode(struct dmi_header *h, u16 ver)
 {
-	struct dmi_header *h=(struct dmi_header *)data;
+	u8 *data=h->data;
 	
 	/*
-	 * Note: DMI types 37, 38 and 39 are untested
+	 * Note: DMI types 37 and 39 are untested
 	 */
 	switch(h->type)
 	{
@@ -2862,11 +2878,19 @@ static void dmi_decode(u8 *data, u16 ver)
 				dmi_string(h, data[0x05]));
 			printf("\tRelease Date: %s\n",
 				dmi_string(h, data[0x08]));
-			printf("\tAddress: 0x%04X0\n",
-				WORD(data+0x06));
-			printf("\tRuntime Size:");
-			dmi_bios_runtime_size((0x10000-WORD(data+0x06))<<4);
-			printf("\n");
+			/*
+			 * On IA-64, the BIOS base address will read 0 because
+			 * there is no BIOS. Skip the base address and the
+			 * runtime size in this case.
+			 */
+			if(WORD(data+0x06)!=0)
+			{
+				printf("\tAddress: 0x%04X0\n",
+					WORD(data+0x06));
+				printf("\tRuntime Size:");
+				dmi_bios_runtime_size((0x10000-WORD(data+0x06))<<4);
+				printf("\n");
+			}
 			printf("\tROM Size: %u kB\n",
 				(data[0x09]+1)<<6);
 			printf("\tCharacteristics:\n");
@@ -2964,7 +2988,7 @@ static void dmi_decode(u8 *data, u16 ver)
 			printf("\tOEM Information: 0x%08X\n",
 				DWORD(data+0x0D));
 			if(h->length<0x15) break;
-			printf("\tHeigth:");
+			printf("\tHeight:");
 			dmi_chassis_height(data[0x11]);
 			printf("\n");
 			printf("\tNumber Of Power Cords:");
@@ -3027,6 +3051,15 @@ static void dmi_decode(u8 *data, u16 ver)
 				dmi_string(h, data[0x21]));
 			printf("\tPart Number: %s\n",
 				dmi_string(h, data[0x22]));
+			if(h->length<0x28) break;
+			if(data[0x23]!=0)
+				printf("\tCore Count: %u\n", data[0x23]);
+			if(data[0x24]!=0)
+				printf("\tCore Enabled: %u\n", data[0x24]);
+			if(data[0x25]!=0)
+				printf("\tThread Count: %u\n", data[0x25]);
+			printf("\tCharacteristics:");
+			dmi_processor_characteristics(WORD(data+0x26), "\t\t");
 			break;
 		
 		case 5: /* 3.3.6 Memory Controller Information */
@@ -3795,12 +3828,25 @@ static void dmi_decode(u8 *data, u16 ver)
 			break;
 		
 		default:
+			if(dmi_decode_oem(h))
+				break;
+			if(opt.flags & FLAG_QUIET)
+				return;
 			printf("%s Type\n",
 				h->type>=128?"OEM-specific":"Unknown");
 			dmi_dump(h, "\t");
 	}
+	printf("\n");
 }
 		
+static void to_dmi_header(struct dmi_header *h, u8 *data)
+{
+	h->type=data[0];
+	h->length=data[1];
+	h->handle=WORD(data+2);
+	h->data=data;
+}
+
 static void dmi_table(u32 base, u16 len, u16 num, u16 ver, const char *devmem)
 {
 	u8 *buf;
@@ -3825,24 +3871,45 @@ static void dmi_table(u32 base, u16 len, u16 num, u16 ver, const char *devmem)
 	}
 	
 	data=buf;
-	while(i<num && data+sizeof(struct dmi_header)<=buf+len)
+	while(i<num && data+4<=buf+len) /* 4 is the length of an SMBIOS structure header */
 	{
 		u8 *next;
-		struct dmi_header *h=(struct dmi_header *)data;
-		int display=((opt.type==NULL || opt.type[h->type])
-			&& !((opt.flags & FLAG_QUIET) && h->type>39)
+		struct dmi_header h;
+		int display;
+
+		to_dmi_header(&h, data);
+		display=((opt.type==NULL || opt.type[h.type])
+			&& !((opt.flags & FLAG_QUIET) && (h.type>39 && h.type<=127))
 			&& !opt.string);
 
+		/*
+		 * If a short entry is found (less than 4 bytes), not only it
+		 * is invalid, but we cannot reliably locate the next entry.
+		 * Better stop at this point, and let the user know his/her
+		 * table is broken.
+		 */
+		if(h.length<4)
+		{
+			printf("Invalid entry length (%u). DMI table is "
+			       "broken! Stop.\n\n", (unsigned int)h.length);
+			opt.flags |= FLAG_QUIET;
+			break;
+		}
+
 		/* In quiet mode, stop decoding at end of table marker */
-		if((opt.flags & FLAG_QUIET) && h->type==127)
+		if((opt.flags & FLAG_QUIET) && h.type==127)
 			break;
 		
 		if(display && !(opt.flags & FLAG_QUIET))
 			printf("Handle 0x%04X, DMI type %d, %d bytes\n",
-				HANDLE(h), h->type, h->length);
+				h.handle, h.type, h.length);
 		
+		/* assign vendor for vendor-specific decodes later */
+		if(h.type==0 && h.length>=5)
+			dmi_set_vendor(dmi_string(&h, data[0x04]));
+
 		/* look for the next handle */
-		next=data+h->length;
+		next=data+h.length;
 		while(next-buf+1<len && (next[0]!=0 || next[1]!=0))
 			next++;
 		next+=2;
@@ -3851,17 +3918,19 @@ static void dmi_table(u32 base, u16 len, u16 num, u16 ver, const char *devmem)
 			if(next-buf<=len)
 			{
 				if(opt.flags & FLAG_DUMP)
-					dmi_dump(h, "\t");
+				{
+					dmi_dump(&h, "\t");
+					printf("\n");
+				}
 				else
-					dmi_decode(data, ver);
+					dmi_decode(&h, ver);
 			}
-			else
-				printf("\t<TRUNCATED>\n");
-			printf("\n");
+			else if(!(opt.flags & FLAG_QUIET))
+				printf("\t<TRUNCATED>\n\n");
 		}
 		else if(opt.string!=NULL
-		     && opt.string->type==h->type
-		     && opt.string->offset<h->length)
+		     && opt.string->type==h.type
+		     && opt.string->offset<h.length)
 		{
 			if (opt.string->lookup!=NULL)
 				printf("%s\n", opt.string->lookup(data[opt.string->offset]));
@@ -3870,7 +3939,7 @@ static void dmi_table(u32 base, u16 len, u16 num, u16 ver, const char *devmem)
 				printf("\n");
 			}
 			else
-				printf("%s\n", dmi_string(h, data[opt.string->offset]));
+				printf("%s\n", dmi_string(&h, data[opt.string->offset]));
 		}
 		
 		data=next;
@@ -3909,7 +3978,6 @@ static int smbios_decode(u8 *buf, const char *devmem)
 	return 0;
 }
 
-#ifndef USE_EFI
 static int legacy_decode(u8 *buf, const char *devmem)
 {
 	if(checksum(buf, 0x0F))
@@ -3924,18 +3992,60 @@ static int legacy_decode(u8 *buf, const char *devmem)
 	
 	return 0;
 }
-#endif /* USE_EFI */
+
+/*
+ * Probe for EFI interface
+ */
+#define EFI_NOT_FOUND   (-1)
+#define EFI_NO_SMBIOS   (-2)
+static int address_from_efi(size_t *address)
+{
+	FILE *efi_systab;
+	const char *filename;
+	char linebuf[64];
+	int ret;
+
+	*address=0; /* Prevent compiler warning */
+
+	/*
+	 * Linux up to 2.6.6: /proc/efi/systab
+	 * Linux 2.6.7 and up: /sys/firmware/efi/systab
+	 */
+	if((efi_systab=fopen(filename="/sys/firmware/efi/systab", "r"))==NULL
+	&& (efi_systab=fopen(filename="/proc/efi/systab", "r"))==NULL)
+	{
+		/* No EFI interface, fallback to memory scan */
+		return EFI_NOT_FOUND;
+	}
+	ret=EFI_NO_SMBIOS;
+	while((fgets(linebuf, sizeof(linebuf)-1, efi_systab))!=NULL)
+	{
+		char *addrp=strchr(linebuf, '=');
+		*(addrp++)='\0';
+		if(strcmp(linebuf, "SMBIOS")==0)
+		{
+			*address=strtoul(addrp, NULL, 0);
+			if(!(opt.flags & FLAG_QUIET))
+				printf("# SMBIOS entry point at 0x%08lx\n",
+				       (unsigned long)*address);
+			ret=0;
+			break;
+		}
+	}
+	if(fclose(efi_systab)!=0)
+		perror(filename);
+
+	if(ret==EFI_NO_SMBIOS)
+		fprintf(stderr, "%s: SMBIOS entry point missing\n", filename);
+	return ret;
+}
 
 int main(int argc, char * const argv[])
 {
 	int ret=0;                  /* Returned value */
 	int found=0;
 	size_t fp;
-#ifdef USE_EFI
-	FILE *efi_systab;
-	const char *filename;
-	char linebuf[64];
-#endif /* USE_EFI */
+	int efi;
 	u8 *buf;
 	
 	if(sizeof(u8)!=1 || sizeof(u16)!=2 || sizeof(u32)!=4 || '\0'!=0)
@@ -3969,37 +4079,15 @@ int main(int argc, char * const argv[])
 	if(!(opt.flags & FLAG_QUIET))
 		printf("# dmidecode %s\n", VERSION);
 	
-#ifdef USE_EFI
-	/*
-	 * Linux up to 2.6.6-rc2: /proc/efi/systab
-	 * Linux 2.6.6-rc3 and up: /sys/firmware/efi/systab
-	 */
-	if((efi_systab=fopen(filename="/proc/efi/systab", "r"))==NULL
-	&& (efi_systab=fopen(filename="/sys/firmware/efi/systab", "r"))==NULL)
+	/* First try EFI (ia64, Intel-based Mac) */
+	efi=address_from_efi(&fp);
+	switch(efi)
 	{
-		perror(filename);
-		ret=1;
-		goto exit_free;
-	}
-	fp=0;
-	while((fgets(linebuf, sizeof(linebuf)-1, efi_systab))!=NULL)
-	{
-		char* addr=memchr(linebuf, '=', strlen(linebuf));
-		*(addr++)='\0';
-		if(strcmp(linebuf, "SMBIOS")==0)
-		{
-			fp=strtoul(addr, NULL, 0);
-			if(!(opt.flags & FLAG_QUIET))
-				printf("# SMBIOS entry point at 0x%08lx\n", fp);
-		}
-	}
-	if(fclose(efi_systab)!=0)
-		perror(filename);
-	if(fp==0)
-	{
-		fprintf(stderr, "%s: SMBIOS entry point missing\n", filename);
-		ret=1;
-		goto exit_free;
+		case EFI_NOT_FOUND:
+			goto memory_scan;
+		case EFI_NO_SMBIOS:
+			ret=1;
+			goto exit_free;
 	}
 
 	if((buf=mem_chunk(fp, 0x20, opt.devmem))==NULL)
@@ -4010,9 +4098,10 @@ int main(int argc, char * const argv[])
 	
 	if(smbios_decode(buf, opt.devmem))
 		found++;
-	
-	free(buf);
-#else /* USE_EFI */
+	goto done;
+
+memory_scan:
+	/* Fallback to memory scan (x86, x86_64) */
 	if((buf=mem_chunk(0xF0000, 0x10000, opt.devmem))==NULL)
 	{
 		ret=1;
@@ -4034,8 +4123,8 @@ int main(int argc, char * const argv[])
 		}
 	}
 	
+done:
 	free(buf);
-#endif /* USE_EFI */
 	
 	if(!found && !(opt.flags & FLAG_QUIET))
 		printf("# No SMBIOS nor DMI entry point found, sorry.\n");
